#![feature(llvm_asm, lang_items, core_intrinsics)]
#![no_std]
#![no_main]

/*
PORTC{0..5} Analog
PORTC6 Reset

PORTB{0..5} Digital {8..13}
PORTD{0..7} Digital {0..7}

PORTB5 LED_BUILTIN
*/

extern crate avr_std_stub;
extern crate ruduino;

// use core::ptr::{read, read_volatile, write_volatile};
use ruduino::cores::current as avr_core;
use ruduino::legacy::serial;
use ruduino::Register;

use avr_core::{DDRB, DDRC, DDRD, PINB, PINC, PIND, PORTB, PORTC, PORTD};

const BLACK: u8 = 0u8;
const RED: u8 = 1u8;
const GREEN: u8 = 2u8;
const YELLOW: u8 = 3u8;

const CPU_FREQUENCY_HZ: u64 = 16_000_000;
const BAUD: u64 = 9600;
const UBRR: u16 = (CPU_FREQUENCY_HZ / 16 / BAUD - 1) as u16;

const HIGH: u8 = 0xFF;
const LOW: u8 = 0x00;

const OUTPUT: u8 = 0xFF;
const INPUT: u8 = 0x00;

#[derive(Clone, PartialEq)]
struct Position {
    x: i8,
    y: i8,
}

#[derive(Clone, PartialEq)]
struct Size {
    width: i8,
    height: i8,
}

#[derive(PartialEq)]
enum Direction {
    Up,
    Down,
    Left,
    Right,
}

#[derive(Clone, PartialEq)]
struct Player {
    position: Position,
    size: Size,
    force: i8,
    speed: f32,
    powered: bool,
    jumping: bool,
}
impl Player {
    fn new() -> Self {
        Player {
            position: Position { x: 2, y: 30 },
            size: Size {
                width: 3,
                height: 6,
            },
            force: 0,
            speed: 0.0,
            powered: false,
            jumping: true,
        }
    }

    fn collide(&self, wall: &Wall) -> Option<Direction> {
        let is_colliding = self.position.x < wall.position.x + wall.size.width
            && self.position.x + self.size.width > wall.position.x
            && self.position.y < wall.position.y + wall.size.height
            && self.position.y + self.size.height > wall.position.y;

        if is_colliding {
            if self.position.y < wall.position.y {
                Some(Direction::Up)
            } else if self.position.x < wall.position.x {
                Some(Direction::Left)
            } else if self.position.x + self.size.width > wall.position.x + wall.size.width {
                Some(Direction::Right)
            } else {
                Some(Direction::Down)
            }
        } else {
            None
        }
    }
}

#[derive(Clone, PartialEq)]
enum WallType {
    Solid,
    Breakable,
}

#[derive(Clone)]
struct Wall {
    position: Position,
    size: Size,
    wall_type: WallType,
    visible: bool,
}

impl Wall {
    fn new(size: Size, position: Position, wall_type: WallType) -> Self {
        Wall {
            position: position,
            size: size,
            wall_type: wall_type,
            visible: true,
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn main() -> ! {
    serial_start();
    DDRB::set_mask_raw(HIGH);
    PORTB::unset_mask_raw(HIGH);
    DDRC::set_mask_raw(HIGH);
    PORTC::unset_mask_raw(HIGH);
    DDRD::set_mask_raw(HIGH);
    PORTD::unset_mask_raw(HIGH);

    let mut step = 0;
    let mut player = Player::new();
    let mut walls: [Wall; 8] = [
        Wall::new(
            Size {
                width: 120,
                height: 5,
            },
            Position { x: 0, y: 54 },
            WallType::Solid,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 50, y: 35 },
            WallType::Solid,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 70, y: 35 },
            WallType::Breakable,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 80, y: 35 },
            WallType::Breakable,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 90, y: 35 },
            WallType::Breakable,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 80, y: 20 },
            WallType::Solid,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 85, y: 35 },
            WallType::Solid,
        ),
        Wall::new(
            Size {
                width: 5,
                height: 5,
            },
            Position { x: 75, y: 35 },
            WallType::Solid,
        ),
    ];

    clear();
    let mut old_player = player.clone();
    let mut last_wall_collisioned: Option<Wall> = None;
    for wall in walls.iter() {
        if wall.visible {
            draw_wall(wall);
        }
    }

    loop {
        // clear();
        draw_player(&player, &old_player);
        old_player = player.clone();

        if player.position.y > 70 {
            player = Player::new();
            continue;
        }

        let mut fall = true;

        // Static wall behavior
        let mut horizontal_collision: Option<Direction> = None;
        for wall in &mut walls {
            if wall.visible {
                match player.collide(wall) {
                    Some(Direction::Up) => {
                        player.jumping = false;
                        player.force = 0;
                        while !player.collide(&wall).is_none() {
                            player.position.y -= 1;
                        }
                        player.position.y += 1;
                        fall = false;
                        draw_wall(&wall);
                    }
                    Some(Direction::Down) => {
                        player.force = 0;
                        player.position.y += 1;
                        fall = true;
                        if wall.wall_type == WallType::Breakable {
                            delete_wall(&wall);
                            wall.visible = false;
                        } else {
                            draw_wall(&wall);
                        }
                    }
                    Some(dir) => {
                        while !player.collide(&wall).is_none() {
                            if dir == Direction::Left {
                                player.position.x -= 1;
                            } else {
                                player.position.x += 1;
                            }
                        }
                        if dir == Direction::Left {
                            player.position.x += 1;
                        } else {
                            player.position.x -= 1;
                        }
                        horizontal_collision = Some(dir);
                        draw_wall(&wall);
                    }
                    // None => fall = true,
                    None => {}
                }
            }
        }

        // Gravity control
        // if the variable `fall` is true it's because all the logic of
        // The platforms doesn't stop the fallen of the character
        if fall {
            step = (step + 1) % 4;
            if step == 0 {
                player.force += 1;
            }
        } else {
            step = 0;
        }
        player.position.y += player.force;
        // player.position.x += 2;

        let values = PINB::read();
        let left = (values >> 0) % 2 == 1;
        let right = (values >> 1) % 2 == 1;
        let jump = (values >> 2) % 2 == 1;

        if left {
            player.position.x += 1;
        }
        if right {
            player.position.x -= 1;
        }
        if jump && !player.jumping {
            player.position.y -= 1;
            player.force = -3;
            player.jumping = true;
        }

        let values = PINB::read();
        for n in 0..8 {
            if (values >> n) % 2 == 1 {
                serial::transmit('1' as u8);
            } else {
                serial::transmit('0' as u8);
            }
        }
        // serial::transmit('\n' as u8);
        // let values = PINC::read();
        // for n in 0..8 {
        //     if (values >> n) % 2 == 1 {
        //         serial::transmit('1' as u8);
        //     } else {
        //         serial::transmit('0' as u8);
        //     }
        // }
        // serial::transmit('\n' as u8);
        // let values = PIND::read();
        // for n in 0..8 {
        //     if (values >> n) % 2 == 1 {
        //         serial::transmit('1' as u8);
        //     } else {
        //         serial::transmit('0' as u8);
        //     }
        // }
        // serial::transmit('\n' as u8);

        // delay(20);
    }
}

fn delay(cycles: u32) {
    for _ in 0..(1_600 * cycles) {
        unsafe { llvm_asm!("" :::: "volatile") }
    }
}

fn clear() {
    for b in ['|' as u8, 0, 2].iter() {
        serial::transmit(*b);
    }
}

fn draw_player(player: &Player, old_player: &Player) {
    if player == old_player {
        return;
    }
    for b in [
        '|' as u8,
        2,
        old_player.position.x as u8,
        old_player.position.y as u8,
        old_player.size.width as u8,
        old_player.size.height as u8 - 1,
        GREEN,
    ]
    .iter()
    {
        serial::transmit(*b);
    }
    for b in [
        '|' as u8,
        2,
        player.position.x as u8,
        player.position.y as u8,
        player.size.width as u8,
        player.size.height as u8 - 1,
        RED,
    ]
    .iter()
    {
        serial::transmit(*b);
    }
}

fn draw_wall(wall: &Wall) {
    for b in [
        '|' as u8,
        2,
        wall.position.x as u8,
        wall.position.y as u8,
        wall.size.width as u8,
        wall.size.height as u8,
        BLACK,
    ]
    .iter()
    {
        serial::transmit(*b);
    }
    if wall.wall_type == WallType::Breakable {
        for b in [
            '|' as u8,
            1,
            wall.position.x as u8 + 2,
            wall.position.y as u8 + 2,
            GREEN,
        ]
        .iter()
        {
            serial::transmit(*b);
        }
    }
}

fn delete_wall(wall: &Wall) {
    for b in [
        '|' as u8,
        2,
        wall.position.x as u8,
        wall.position.y as u8,
        wall.size.width as u8,
        wall.size.height as u8,
        GREEN,
    ]
    .iter()
    {
        serial::transmit(*b);
    }
}

fn serial_start() {
    serial::Serial::new(UBRR)
        .character_size(serial::CharacterSize::EightBits)
        .mode(serial::Mode::Asynchronous)
        // .mode(serial::Mode::MasterSpi)
        .parity(serial::Parity::Disabled)
        .stop_bits(serial::StopBits::OneBit)
        .configure();
}

// These do not need to be in a module, but we group them here for clarity.
// pub mod std {
// #[lang = "eh_personality"]
// pub unsafe extern "C" fn rust_eh_personality(
// _state: (),
// _exception_object: *mut (),
// _context: *mut (),
// ) -> () {
// }
//
// #[panic_handler]
// fn panic(_info: &::core::panic::PanicInfo) -> ! {
// loop {}
// }
// }
